/*
 * The MIT License
 *
 * Copyright 2017 Germán Bordel <german.bordel at ehu.es>.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package edu.upvehu.gbg.ejemplos.pvdb;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

/**
 * Clase asociada al alineamiento voz/acta de una sesión del Parlamento Vasco
 * @author Germán Bordel &lt;german.bordel at ehu.es&gt;
 */
public class PVSession {
    private List<Palabra> palabras;
    private List<Fonema> fonemas;
    private Estadisticas estadisticas;
    
    /**
     * @param url dirección web del fichero de texto de alineamiento
     * @throws MalformedURLException la URL es incorrecta
     * @throws IOException problemas de acceso al fichero remoto.
     */
    public PVSession(String url) throws MalformedURLException, IOException {
        fonemas=new ArrayList<>();
        palabras=new ArrayList<>();
        BufferedReader br = new BufferedReader(new InputStreamReader(new URL(url).openStream()));
        String linea;
        for (int i=0, startIndex=0;(linea=br.readLine())!=null;) {
            linea=linea.trim();
            if (linea.length()==0 || linea.startsWith("#")) continue;
            
            String[] fonrealFonrecoInfo=linea.split("\t");
            String[] info=fonrealFonrecoInfo[2].trim().split("\\|");
            double start=0,end=0,prob=0;
            try {
                String[] numeros=info[2].trim().split("\\s+");
                start=Double.parseDouble(numeros[0]);
                end=Double.parseDouble(numeros[1]);
                prob=Double.parseDouble(numeros[2]);
            } catch(ArrayIndexOutOfBoundsException ignore) {}
            fonemas.add(new Fonema(fonrealFonrecoInfo[0].charAt(0),fonrealFonrecoInfo[1].charAt(0),start,end,prob));
            if (info.length>=2 && info[1].length()>0 && i>0) {
                palabras.add(new Palabra(cleanWord(info[1]),startIndex,i-1));
                startIndex=i;
            }
            i++;
        }
    }
    
    private static String cleanWord(String w) {return w.replaceAll("\\.|,|'|\"", "");}
    public List<Palabra> getPalabras() {return palabras;}
    public List<Fonema> getFonemas() {return fonemas;}
    
    /**
     * Retorna las estadísticas de la sesión.
     * Si no han sido calculadas previamente, realiza los calculos. En caso contrario devuelve los precalculados.
     * @return las esatdísticas de la sesión
     */
    Estadisticas stats(){
        if (estadisticas!=null) return estadisticas; 
        
        double media=0;
        int validas=0;
        for (Palabra p:palabras) {
            double inicio=fonemas.get(p.inicio).start, fin=fonemas.get(p.fin).end;
            if (inicio==0 || fin==0) continue; 
            media+=fin-inicio;
            validas++;
        }
        media/=validas;

        double desviacionTipica=0;
        for (Palabra p:palabras) {
            double inicio=fonemas.get(p.inicio).start, fin=fonemas.get(p.fin).end;
            if (inicio==0 || fin==0) continue; 
            desviacionTipica+=(fin-inicio-media)*(fin-inicio-media);
        }
        desviacionTipica=Math.sqrt(desviacionTipica/validas);
        return estadisticas=new Estadisticas(palabras.size(),validas,media, desviacionTipica);
    }

    /**
     * Clase que almacena la información de alineamiento de un fonema
     */    
    public static class Fonema {
        char real,reco;
        double start,end,prob;

        public Fonema(char real, char reco, double start, double end, double prob) {
            this.real = real; this.reco = reco; this.start = start; this.end = end; this.prob = prob;
        }
     }

    /**
     * Clase que asocia cada palabra a la secuencia de fonemas alineados
     */    
    public static class Palabra {
        String palabra;
        int inicio,fin;

            public Palabra(String palabra, int inicio, int fin) {
                this.palabra = palabra; this.inicio = inicio; this.fin = fin;
            }
    }

    /**
     * Clase que agrupa las estadísticas de la sesión.
     * "Sabe" llevar sus datos a una base de datos a traves de la conexión que se le aporte.
     */
    public static class Estadisticas {
        int numPalabras, validas;
        double media, desviacionTipica;

        public Estadisticas(int numPalabras, int validas, double media, double desviacionTipica) {
            this.numPalabras=numPalabras; this.validas=validas; this.media = media; this.desviacionTipica = desviacionTipica;
        }
        
        void dump2DB(Connection conexion, String id) throws SQLException{
            conexion.createStatement().executeUpdate(
                String.format(Locale.US,"insert into estadisticas (id, numPalabras, validas, media,desviacionTipica) values (%s, %d, %d, %.3f, %.3f)",id, numPalabras,validas, media,desviacionTipica)
            );
        }
    }
}